# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_front_end_common.utilities.iobuf_extractor import IOBufExtractor


def chip_io_buf_extractor():
    """
    Extract the logging output buffers from the machine, and separates
    lines based on their prefix.

    :return: error_entries, warn_entries
    :rtype: tuple(list(str),list(str))
    """
    extractor = IOBufExtractor()
    return extractor.extract_iobuf()
